"""Contains all the standard gates that can be used in the circuit.
These include:

- :class:`~qclight.gate.h_gate.HGate`
- :class:`~qclight.gate.x_gate.XGate`
- :class:`~qclight.gate.z_gate.ZGate`
- :class:`~qclight.gate.i_gate.IGate`
"""
from .gate import Gate
from .h_gate import HGate
from .x_gate import XGate
from .z_gate import ZGate
from .i_gate import IGate
